// scalar product: 3: recursive local summation and summation on GPU
//                    includes special implementation for <x,x>
// Profiler:  export CUDA_PROFILE=1
#include <cassert>
#include <ctime>
#include <iostream>
#include <string>
using namespace std;

/** Prints GPU info into the terminal.
*/
void printGPUInfo()
{
    // https://devblogs.nvidia.com/how-query-device-properties-and-handle-errors-cuda-cc/
    int nDevices;
    cudaError_t err = cudaGetDeviceCount(&nDevices);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));
    
    //  https://docs.nvidia.com/cuda/cuda-runtime-api/structcudaDeviceProp.html 
    cudaDeviceProp prop;
    // https://www.cs.cmu.edu/afs/cs/academic/class/15668-s11/www/cuda-doc/html/group__CUDART__DEVICE_g5aa4f47938af8276f08074d09b7d520c.html
    err = cudaGetDeviceProperties (&prop, 0);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));
    cout << endl << "We work on " << nDevices << " x " << prop.name << " GPU  with "
         << prop.multiProcessorCount << " Multiprocessors"
         << ", Compute capability " << prop.major << "." << prop.minor << endl;
    cout << "global Mem: " << prop.totalGlobalMem/1024/1000/1000 << " GB" 
         << "   shared Mem per SME: " << prop.sharedMemPerMultiprocessor/1024 << "kB"
         << "   shared Mem per Block: " << prop.sharedMemPerBlock/1024 << " kB" <<endl;
    cout << "global L2 cache: " << prop.l2CacheSize/1024 << " kB"
         << "    local L1 cache supported: " << boolalpha << bool(prop.localL1CacheSupported)  << endl;
    cout << "max Threads per Block:" << prop.maxThreadsPerBlock << endl;
}

/** 	Device function. 
    Calculates the inner product per thread block
	@param[out] sg	vector on device storing the partial inner products
	@param[in] x	vector on device
	@param[in] y	vector on device
	@param[in] N	number of vector elements
    @warning shared memory of blocksize * sizeof(double) bytes has to be requested on kernel call
*/
__global__ static void scalar(double *sg, const double *x, const double *y, int N)
{
    extern __shared__ double sdata[];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int str = gridDim.x * blockDim.x;

    double ss = 0.0;
    for (int i = idx; i < N; i += str)
        ss += x[i] * y[i];
    sdata[threadIdx.x] = ss;

    __syncthreads();
    for (unsigned int s = blockDim.x >> 1; s > 0; s >>= 1)
    {
        if (threadIdx.x < s)
        {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }
    if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}

/** 	Device function. 
    Calculates the square of the L2 norm per thread block
	@param[out] sg	vector on device storing the partial inner products
	@param[in] x	vector on device
	@param[in] N	number of vector elements
    @warning shared memory of blocksize * sizeof(double) bytes has to be requested on kernel call
*/
__global__ static void norm2(double *sg, const double *x, int N)
{
    extern __shared__ double sdata[];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int str = gridDim.x * blockDim.x;

    double ss = 0.0;
    for (int i = idx; i < N; i += str)
        ss += x[i] * x[i];
    sdata[threadIdx.x] = ss;

    __syncthreads();
    for (unsigned int s = blockDim.x >> 1; s > 0; s >>= 1)
    {
        if (threadIdx.x < s)
        {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }
    if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}

/** 	Device function. 
    Kernel function for addition on  o n e  block on device
	@param[inout] s	vector on device , its sum is stored in @p s[0]
	@param[in] N	number of vector elements
    @warning shared memory of blocksize * sizeof(double) bytes has to be requested on kernel call
    @see scalar
*/
__global__ static void add_1_block(double *s, int N)
{
    if (N > static_cast<int>(blockDim.x)) return;
    extern __shared__ double sdata[];
    const int tid = threadIdx.x;
    if (tid < N) sdata[tid] = s[tid];
    else       sdata[tid] = 0.0;

    __syncthreads();
    for (int stride = blockDim.x / 2; stride > 0; stride >>= 1)
    {
        if (tid < stride)
        {
            sdata[tid] += sdata[tid + stride];
        }
        __syncthreads();
    }
    if (tid == 0)  s[0] = sdata[0];
}

//----------------------------------------------------------------------
static bool b_init = false;             // temp. memory (on GPU) allocated for dscapr_GPU?

/** 	Host function. 
 *  Inner product calculated with device vectors.
 *  The implementation consists of two kernel with a temporary array for data transfer between them.
 *  The necessary dynamic memory alocation is done only in the first function call (global variable @p b_init).
 *  This  dynamic memory is not deallocated afterwards.
	@param[in] N	number of vector elements
	@param[in] x_d	vector on device
	@param[in] y_d	vector on device
	@return 	resulting Euclidian inner product <x_d,y_d> on host
	@warning    The dynamic memory is not deallocated afterwards.
*/
double static dscapr_GPU(const int N, const double x_d[], const double y_d[])
{
    static double *s_d;                 // allocated temp. global memory
    static int sBytes;                  //              size of that in Byte
    static dim3 dimBlock, dimGrid;      // #blocks, #threads per block to start
    if (!b_init)  
    {                          // This scope is exexute only in the first function call
        b_init = !b_init;      //            prevents a second execution
        //  https://docs.nvidia.com/cuda/cuda-runtime-api/structcudaDeviceProp.html 
        cudaDeviceProp prop;
        cudaGetDeviceProperties (&prop, 0); // get device properties

        const int blocksize = 8 * 64,
                  gridsize = prop.multiProcessorCount;
        sBytes   = blocksize * sizeof(double);
        dimBlock = dim3(blocksize);     // #threads per block
        dimGrid  = dim3(4*gridsize);    // #blocks

        cudaMalloc((void **) &s_d, blocksize * sizeof(double)); // temp. global memory on device
        cout << "\n###  Fast version of inner product (incl. norm2) on device is used. ###\n";
        cout << "     " << dimGrid.x << " x " << dimBlock.x << " Threads\n\n";
    }
   
// call the kernel function with  dimGrid.x * dimBlock.x threads and shared array
    if (x_d!=y_d)
    {                         //        |--> sdata[sBytes]
        scalar <<< dimGrid, dimBlock, sBytes>>>(s_d, x_d, y_d, N);
    }
    else  // x_d and y_d are the same vectors
    {                         //        |--> sdata[sBytes]
        norm2 <<< dimGrid, dimBlock, sBytes>>>(s_d, x_d, N);
    }
//    power of 2 as number of threads per block
    const unsigned int oneBlock = 1 << (int)ceil(log(dimGrid.x + 0.0) / log(2.0));
    add_1_block <<< 1, oneBlock, oneBlock *sizeof(double)>>>(s_d, dimGrid.x);
// copy data:  device --> host
    double sum_h;
    cudaMemcpy(&sum_h, s_d, sizeof(double), cudaMemcpyDeviceToHost);

    //cudaFree(s_d);                   // GH: ToDo: Free memory after the end of all computations
    return sum_h;
}

//----------------------------------------------------------------------
int main(int argc, char **argv )       // main function on host
{
    printGPUInfo();
    
    // -------- chose GPU to use ------------------
    int nDevices;
    cudaGetDeviceCount(&nDevices);      // #GPUs in computer
    int const chooseGPU=nDevices-1;
    //int const chooseGPU=0;
    cudaSetDevice(chooseGPU);           // choose last GPU, ie. 3 ot of 4
    cout << "## device no " << chooseGPU << " is used." << endl;
    // --------------------------------------------
    
    const int LOOP = 100;               // #function calls
    int N = 100000000;                  // problem size
    if (argc>1) N = stoi(string(argv[1]));  // allows the flexible call ./main.NVCC_ 200000000
    cout << "\nN = " << N << endl;
    const int nBytes = N * sizeof(double);
    cout << 2*nBytes/1024/1024 << " MB to allocate." << endl;

    double *x_h, *y_h;                  // host data
    x_h = new double [N];               // allocate on host
    y_h = new double [N];

    double *x_d, *y_d;                  // device data
    cudaMalloc((void **) &x_d, nBytes); // allocate on device
    cudaMalloc((void **) &y_d, nBytes);

    double ss = 0.0;
    for (int i = 0; i < N; i++)
    {
        x_h[i] = (i % 137) + 1;         // init data on host
        y_h[i] = 1.0 / x_h[i];
        ss += x_h[i] * y_h[i];
    }
                                        // copy data:  host --> device
    cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice);
    cudaMemcpy(y_d, y_h, nBytes, cudaMemcpyHostToDevice);

    double sum = dscapr_GPU(N, x_d, y_d); // Dummy call to set up temp. memory

    // ---------- start benchmark ------------------
    cudaDeviceSynchronize();
    auto tstart = clock();
    for (int k = 0; k < LOOP; ++k)
    {
        sum += dscapr_GPU(N, x_d, y_d);
    }
    cudaDeviceSynchronize();
    double tgpu = static_cast<double>(clock() - tstart) / CLOCKS_PER_SEC / LOOP;
    // ---------- end benchmark ------------------

    cout << "GPU: " << tgpu << " sec.; " <<
         2e-9 * N / tgpu << " GFLOPS;  " <<
         2.0 * N *sizeof(double) / tgpu / 1024 / 1024 / 1024 << " GiB/sec.\n ";

    if (static_cast<int>(sum/(LOOP+1)) != N) // check incorrectness of result
    {
        cout << " !!  W R O N G  result  " << endl;
        cout << sum << "  vs.  " << N << "(" << ss << ")" << endl;
    }
    else
        cout << " correct  result  " << sum/(LOOP+1) << endl;

    delete [] x_h; delete [] y_h;
    cudaFree(x_d); cudaFree(y_d);
    return 0;
}
