// scalar product: 4: use of cublasDdot
#include <cassert>
#include <ctime>
#include <cublas_v2.h>   //  new API, see https://docs.nvidia.com/cuda/cublas/index.html#new-and-legacy-cublas-api
#include <iostream>
using namespace std;

int main(void)         //  main function on host
{
    const int N = 100000000;
    //const int N = 14000000;
    const int LOOP = 100;

    double *x_h = new double [N]; // host data
    double *y_h = new double [N]; // host data

    double ss = 0.0;
    for (int i = 0; i < N; i++)
    {
        x_h[i] = (i % 137) + 1;
        y_h[i] = 1.0 / x_h[i];
        ss += x_h[i] * y_h[i];
    }
//  -- Initialize cuBLAS, check GPU
//  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
    //cudaError_t  cudaStat;               //  cudaMalloc  status
    cublasStatus_t  stat;                //  CUBLAS  functions  status
    cublasHandle_t  handle;

    stat = cublasCreate (& handle );   //  initialize  CUBLAS  context
    if (stat != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "ERROR: cublasInit() failed!\n");
        exit(1);
    }

    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
         prop.multiProcessorCount << " Multiprocessors" << endl;

//  -- Move data from host to device
// allocate memory on device
    double *x_d, *y_d; // device data
    cudaMalloc((void **) &x_d, N * sizeof(double));
    cudaMalloc((void **) &y_d, N * sizeof(double));

// copy data:  host --> device
    cublasSetVector(N, sizeof(double), x_h, 1, x_d, 1);
    cublasSetVector(N, sizeof(double), y_h, 1, y_d, 1);

//  -- Benchmark
    double sum = 0.0;
    cudaDeviceSynchronize();
    auto tstart = clock();
    for (int k = 0; k < LOOP; ++k)
    {
        double res;
        stat = cublasDdot(handle, N, x_d, 1, y_d, 1, &res);
        sum += res;
    }
    cudaDeviceSynchronize();
    double tgpu = static_cast<double>(clock() - tstart) / CLOCKS_PER_SEC / LOOP;

//  -- Statistics, check result
    cout << "GPU: " << tgpu << " sec.; " <<
         2e-9 * N / tgpu << " GFLOPS;  " <<
         2.0 * N *sizeof(double) / tgpu / 1024 / 1024 / 1024 << " GiB/sec.\n ";

    if (static_cast<int>(sum / LOOP) != N)
    {
        cout << " !!  W R O N G  result  " << endl;
        cout << sum << "  vs.  " << N << "(" << ss << ")" << endl;
    }
    else
        cout << " correct  result  " << endl;

//  -- Free all cuBlas ressources
    cudaFree(x_d);
    cudaFree(y_d);
    cublasDestroy(handle );

//  -- free host ressources
    delete [] x_h;
    delete [] y_h;

    return 0;
}
